/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

#include "common.h"

#include <errno.h>
#include <pwd.h>
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif

#include "lock.h"
#include "adm.h"
#include "adm_add.h"
#include "udb_check.h"
#include "set_uid.h"

#if defined(UD_DBM) || defined(AUTH_DBM) || defined(UID_MAP_FILE)
#include "hashfile.h"
#endif



/* ADDUDB - Add an entry to the user database describing a user with
 * the given login, uid, gid, password, fullname, and directory.
 */

#if defined(UD_PASSWD) && defined(PASSWD_FILE)

static
void addudb(char *lid, gid_t gid, uid_t uid, char *pw, char *fn, char *dir,
    int status)
{
char bf[BFSZ];
int fd;
size_t len;
char *encpw;

#ifdef SHARE_UPW
    encpw= mkpasswd(pw,status);
#else
    encpw= "*";
#endif

    sprintf(bf,"%s:%s:%d:%d:%s:%s:\n", lid, encpw, uid, gid, fn, dir);
    len= strlen(bf);
    
    set_httpd_if_shared();
    if ((fd= open(PASSWD_FILE, O_WRONLY|O_APPEND)) < 0)
    {
	set_cfadm_if_shared();
	die("addudb cannot open "PASSWD_FILE" to write");
    }
    lock_exclusive(fd,PASSWD_FILE);

    if (write(fd, bf, len) != len)
    {
	set_cfadm_if_shared();
	die("write to "PASSWD_FILE" failed");
    }
    
    close(fd);
    unlock(PASSWD_FILE);

    set_cfadm_if_shared();
}

#endif /* UD_PASSWD && PASSWD_FILE */

#ifdef UD_DBM

static
void addudb(char *lid, gid_t gid, uid_t uid, char *pw, char *fn, char *dir,
    int status)
{
char bf[BFSZ];
char *encpw;

#ifdef SHARE_UPW
    encpw= mkpasswd(pw,status);
#else
    encpw= "*";
#endif

    set_httpd_if_shared();

#ifdef DBM_NAM_FILE

    sprintf(bf,"%s:%d:%d:%s:%s:\n", encpw, uid, gid, fn, dir);

    if (storedbm(DBM_NAM_FILE,
		 lid, strlen(lid),	/* don't include terminating NUL */
		 bf, strlen(bf)+1,	/* include terminating NUL */
		 1))			/* create file if necessary */
    {
	set_cfadm_if_shared();
	die("Cannot open user database "DBM_NAM_FILE" (%s)",errdbm());
    }
#endif

#ifdef DBM_UID_FILE

    sprintf(bf,"*:%s:%d:%s:%s:\n", lid, gid, fn, dir);

    if (storedbm(DBM_UID_FILE,
		 (char *)&uid, sizeof(int),
		 bf, strlen(bf)+1,	/* include terminating NUL */
		 1))			/* create file if necessary */
    {
	set_cfadm_if_shared();
	die("Cannot open user database "DBM_UID_FILE" (%s)",errdbm());
    }
#endif

    set_cfadm_if_shared();
}

#endif /* UD_DBM */

#ifdef UD_TAGFILE

static
void addudb(char *lid, gid_t gid, uid_t uid, char *pw, char *fn, char *dir,
    int status)
{
char fname[BFSZ];
FILE *fp;

    sprintf(fname,"%s/"IDENT_FILE,dir);
    if ((fp= fopen(fname,"w")) == NULL)
	die("could not create ident file %s",fname);
    lock_exclusive(fileno(fp),fname);
    fprintf(fp,"*uid=%d\n*gid=%d\nfullname={%s}\n", uid, gid, fn);
    fclose(fp);
    unlock(fname);
}

#endif /* UD_TAGFILE */


#ifndef SHARE_UPW

/* ADDPWDB - Add an entry to the password file or database for a user.  This
 * should be called only if the password file and user database are separate.
 */

#ifdef AUTH_FILE

static
void addpwdb(char *lid, char *pw, int status)
{
char bf[BFSZ];
int fd;
size_t len;

    sprintf(bf,"%s:%s\n", lid, mkpasswd(pw,status));
    len= strlen(bf);
    
    if ((fd= open(AUTH_FILE, O_WRONLY|O_APPEND)) < 0)
	die("addpwdb cannot open "AUTH_FILE" to write");
    lock_exclusive(fd,AUTH_FILE);

    if (write(fd, bf, len) != len)
	die("write to "AUTH_FILE" failed");

    close(fd);
    unlock(AUTH_FILE);
}

#endif /*AUTH_FILE */

#ifdef AUTH_DBM

static
void addpwdb(char *lid, char *pw, int status)
{
char *data= mkpasswd(pw,status);

    if (storedbm(AUTH_DBM,
    		 lid, strlen(lid),	/* Don't include terminating NUL */
    		 data, strlen(data)+1,	/* Include terminating NUL */
    		 1))			/* Create file if necessary */
	die("Cannot open user database "AUTH_DBM" (%s)",errdbm());
}
#endif /* AUTH_DBM */

#endif /*SHARE_UPW*/


/* MK_DIRECTORY - Create a directory for the user
 */

static void mk_directory(char *dir, char *lid)
{
extern int errno;
char *p;

    strcpy(dir, USER_DIR);
    p= dir + strlen(USER_DIR);

#if USER_DIR_LEVEL > 0

    *p++= lid[0];
    *p= 0;
    if (mkdir(dir,0755) && errno != EEXIST)
	die("Could not create directory %s: %s", dir, strerror(errno));
    *p++= '/';
#if USER_DIR_LEVEL > 1

    *p++= lid[1] == '\0' ? '_' : lid[1];
    *p= 0;
    if (mkdir(dir,0755) && errno != EEXIST)
	die("Could not create directory %s: %s", dir, strerror(errno));
    *p++= '/';
#endif
#endif

    strcpy(p,lid);
    if (mkdir(dir,0755))
	die("Could not create directory %s: %s", dir, strerror(errno));
}


/* NEXT_UID - Return the next available uid number.  If <inc> is true,
 * increment it.
 */

uid_t next_uid(int inc)
{
FILE *uid_fp;
uid_t uid;

    /* Get a lock on the next_uid file */
    if ((uid_fp= fopen(NEXT_UID_FILE,"r+")) == NULL)
	die("cannot open next-uid-file "NEXT_UID_FILE);
    lock_exclusive(fileno(uid_fp),NEXT_UID_FILE);

    /* Get the uid number to use */
    if (fscanf(uid_fp,sizeof(uid_t)==sizeof(short)?"%hd":"%d",&uid) != 1)
    {
	fclose(uid_fp);
	unlock(NEXT_UID_FILE);
	die("cannot read next uid from "NEXT_UID_FILE);
    }

    if (inc)
    {
	/* Increment the uid number file */
	rewind(uid_fp);
	fprintf(uid_fp,"%d\n",uid+1);
    }

    /* Release the file */
    fclose(uid_fp);
    unlock(NEXT_UID_FILE);

    return uid;
}


/* ADDUSER - create an account for a user.  "dir" is an output, the rest are
 * inputs.  Status is the initial status, 0=validated, 1=unvalidated,
 * 2=invalidated (though doing 2 would be wierd).
 */

void adduser(char *lid, gid_t gid, char *pw, char *fn, int status, char *dir)
{
uid_t uid;

    set_cfadm();

    /* Get the next available uid number */
    uid= next_uid(1);

    /* Create user's directory and store name in dir */
    mk_directory(dir,lid);

    /* Create the passwd file entries */
    addudb(lid, gid, uid, pw, fn, dir, status);
#ifndef SHARE_UPW
    set_httpd();
    addpwdb(lid, pw, status);
    set_cfadm();
#endif 
}


#ifdef UID_MAP_FILE

/* LINKUSER - create a link from the given real uid to given conference lid.
 */

void linkuser(uid_t uid, char *lid)
{
    set_cfadm();

    if (storedbm(UID_MAP_FILE,
    		 (char *)&uid, sizeof(uid),
    		 lid, strlen(lid)+1,
    		 1))
	die("unable to open map file "UID_MAP_FILE" (%s)",errdbm());
}

#endif /* UID_MAP_FILE */
